from itertools import product
import re
from typing import Callable

import numpy as np
import pandas as pd

MIN_PRICE: int = 0
MAX_PRICE: int = 300000
MIN_CNT: int = 0
MAX_CNT: int = 500


def cast_to_datetime(df: pd.DataFrame) -> pd.DataFrame:
    df["date"] = pd.to_datetime(df["date"], dayfirst=True)
    return df


def filter_neg_price(df: pd.DataFrame) -> pd.DataFrame:
    return df[df["item_price"] > MIN_PRICE]


def filter_returns(df: pd.DataFrame) -> pd.DataFrame:
    return df[df["item_cnt_day"] > MIN_CNT]


def filter_outliers_price(df: pd.DataFrame) -> pd.DataFrame:
    return df[df["item_price"] < MAX_PRICE]


def filter_outliers_cnt(df: pd.DataFrame) -> pd.DataFrame:
    return df[df["item_cnt_day"] < MAX_CNT]


def filter_duplicates(df: pd.DataFrame) -> pd.DataFrame:
    return df.drop_duplicates()


def clean_text(df: pd.DataFrame) -> pd.DataFrame:
    regex = re.compile("[^+a-zA-Zа-яА-Я0-9]+")
    col_name = [name for name in df.columns if "name" in name][0]
    df[col_name] = df[col_name].apply(lambda text: regex.sub(" ", text.strip().lower()))
    return df


def replace_duplicated_shop_ids(df: pd.DataFrame) -> pd.DataFrame:
    df["shop_id"] = df["shop_id"].replace({0: 57, 1: 58, 10: 11})
    return df


def expand_shop_item_grid(df: pd.DataFrame) -> pd.DataFrame:
    group_dfs = []
    for _, group_df in df.groupby("date_block_num"):
        group_dfs.append(
            pd.DataFrame(
                product(
                    group_df["shop_id"].unique(),
                    group_df["item_id"].unique(),
                    group_df["date_block_num"].unique(),
                ),
                columns=["shop_id", "item_id", "date_block_num"],
            )
        )

    expanded_df = pd.concat(group_dfs, axis=0).merge(
        group_df, how="left", on=["shop_id", "item_id", "date_block_num"]
    )

    return expanded_df


def group_by_month(df: pd.DataFrame) -> pd.DataFrame:
    """Function that groups df by date_block_num,
    expands it with shop/item ids and sorts values by date_block_num"""
    # group by month
    monthly_sales = (
        df.groupby(["date_block_num", "shop_id", "item_id"])
        .agg({"item_price": "mean", "item_cnt_day": "sum"})
        .reset_index()
        .rename(columns={"item_cnt_day": "item_cnt_month"})
    )

    # expand (for every month create a grid from all shops/items combinations from that month)
    monthly_sales = expand_shop_item_grid(monthly_sales)

    monthly_sales["item_cnt_month"] = monthly_sales["item_cnt_month"].fillna(0)
    monthly_sales["item_price"] = monthly_sales["item_price"].fillna(0)
    monthly_sales["item_cnt_month"] = monthly_sales["item_cnt_month"].astype(int)
    # monthly_sales["item_cnt_month"] = monthly_sales["item_cnt_month"].clip(0, 20)

    # sort
    monthly_sales.sort_values(["date_block_num", "shop_id", "item_id"], inplace=True)

    return monthly_sales


def reduce_mem_usage(df: pd.DataFrame, silent: bool = True) -> pd.DataFrame:
    """
    Iterates through all the columns of a dataframe and downcasts the data type
     to reduce memory usage.
    """

    def downcast_numeric(series: pd.Series) -> pd.Series:
        """
        Downcast a numeric series into the smallest possible int/float dtype.
        """
        if pd.api.types.is_integer_dtype(series.dtype):
            series = pd.to_numeric(series, downcast="integer")
        if pd.api.types.is_float_dtype(series.dtype):
            series = pd.to_numeric(series, downcast="float")
        return series

    if not silent:
        start_mem = np.sum(df.memory_usage()) / 1024**2
        print("Memory usage of dataframe is {:.2f} MB".format(start_mem))

    for col in df.columns:
        df.loc[:, col] = downcast_numeric(df.loc[:, col])

    if not silent:
        end_mem = np.sum(df.memory_usage()) / 1024**2
        print("Memory usage after optimization is: {:.2f} MB".format(end_mem))
        print("Decreased by {:.1f}%".format(100 * (start_mem - end_mem) / start_mem))

    return df


transforms: dict[str, list[Callable[[pd.DataFrame], pd.DataFrame]]] = {
    "sales_train": [
        cast_to_datetime,
        filter_neg_price,
        filter_returns,
        filter_outliers_price,
        filter_outliers_cnt,
        filter_duplicates,
        replace_duplicated_shop_ids,
        group_by_month,
        reduce_mem_usage,
    ],
    "test": [
        replace_duplicated_shop_ids,
        reduce_mem_usage,
    ],
    "shops": [
        reduce_mem_usage,
    ],
    "items": [
        reduce_mem_usage,
    ],
    "item_categories": [
        reduce_mem_usage,
    ],
}



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['future_sales', 'future_sales.data']

package_data = \
{'': ['*']}

install_requires = \
['category-encoders>=2.5.1,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'hyperopt>=0.2.7,<0.3.0',
 'pandas>=1.5.2,<2.0.0',
 'scikit-learn>=1.2.0,<2.0.0',
 'xgboost>=1.7.2,<2.0.0']

entry_points = \
{'console_scripts': ['predict = future_sales.predict:predict',
                     'train = future_sales.train:train']}

setup_kwargs = {
    'name': 'future-sales',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Sabina Bayramova',
    'author_email': 'sb.sabina00@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.cluster import AgglomerativeClustering
from sklearn.decomposition import PCA

from future_sales.data.utils import reduce_mem_usage


def extract_shop_features(df: pd.DataFrame, shops: pd.DataFrame) -> pd.DataFrame:
    shops["city"] = shops["shop_name"].apply(lambda shop_name: shop_name.split()[0])
    shops["city"] = shops["city"].replace("!", "", regex=True)
    shops["city_code"] = shops["city"].factorize()[0]

    shops["shop_type"] = shops["shop_name"].apply(
        lambda shop_name: shop_name.split()[1]
        if shop_name.split()[1] in ["ТЦ", "ТРК", "ТРЦ", "ТК"]
        else "-"
    )
    shops["shop_type_code"] = shops["shop_type"].factorize()[0]

    shops["shop_first_month"] = df.groupby("shop_id")["date_block_num"].min()
    shops["shop_last_month"] = df.groupby("shop_id")["date_block_num"].max()

    shops.dropna(axis=0, inplace=True)

    return shops[
        [
            "shop_id",
            "city_code",
            "shop_type_code",
            "shop_first_month",
            "shop_last_month",
        ]
    ]


def extract_item_category_features(item_cats: pd.DataFrame) -> pd.DataFrame:
    item_cats["category_group"] = item_cats["item_category_name"].apply(
        lambda x: x.strip().split("-")[0]
    )
    item_cats["category_subgroup"] = item_cats["item_category_name"].apply(
        lambda x: x.strip().split("-")[1] if len(x.split("-")) > 1 else x.split("-")[0]
    )

    item_cats.loc[
        item_cats["category_group"].str.startswith("Игры"), "category_group"
    ] = "Игры"
    item_cats.loc[
        item_cats["category_group"] == "Карты оплаты (Кино, Музыка, Игры)",
        "category_group",
    ] = "Карты оплаты"
    item_cats.loc[
        item_cats["category_group"] == "Чистые носители (шпиль)", "category_group"
    ] = "Чистые носители"
    item_cats.loc[
        item_cats["category_group"] == "Чистые носители (штучные)", "category_group"
    ] = "Чистые носители"

    item_cats["category_group_code"] = item_cats["category_group"].factorize()[0]
    item_cats["category_subgroup_code"] = item_cats["category_subgroup"].factorize()[0]

    return item_cats[
        ["item_category_id", "category_group_code", "category_subgroup_code"]
    ]


def extract_item_features(df: pd.DataFrame, items: pd.DataFrame) -> pd.DataFrame:
    items["item_first_sale"] = df.groupby("item_id")["date_block_num"].min()
    items["item_first_sale"] = items["item_first_sale"].fillna(34)
    return items[["item_id", "item_category_id", "item_first_sale"]]


def extract_cluster_feature(
    df: pd.DataFrame, cluster_name: str, n_clusters: int
) -> pd.DataFrame:
    # PCA:
    pca = PCA(n_components=2, random_state=42)
    components = pca.fit_transform(df)
    components = pd.DataFrame(components)

    # Clustering:
    clusterer = AgglomerativeClustering(n_clusters=n_clusters, linkage="average")
    labels = clusterer.fit_predict(components)

    df.insert(loc=0, column=cluster_name, value=labels)

    return df


def extract_lagged_features(
    df: pd.DataFrame, lags: list[int], col_name: str
) -> pd.DataFrame:
    for i in lags:
        feature_name = f"{col_name}_lag_{i}"
        shifted = df.copy()
        shifted.rename(columns={"item_cnt_month": feature_name}, inplace=True)
        shifted["date_block_num"] = shifted["date_block_num"] + i
        df = df.merge(
            shifted[["shop_id", "item_id", "date_block_num", feature_name]],
            on=["shop_id", "item_id", "date_block_num"],
            how="left",
        )
        df[feature_name] = df[feature_name].fillna(0)
    return df


def extract_rolling_window_features(
    df: pd.DataFrame, windows: list[int], col_name: str, agg_func: str
) -> pd.DataFrame:
    for w in windows:
        feature_name = f"{col_name}_rolling_{w}_{agg_func}"
        df[feature_name] = df.groupby(by=["shop_id", "item_id"], group_keys=False)[
            col_name
        ].apply(lambda x: x.rolling(window=3).agg(agg_func))
        df[feature_name] = df[feature_name].fillna(0)
    return df


def build_features(
    data: dict[str, pd.DataFrame],
    indices: Union[npt.NDArray[np.int_], list[int]],
    is_test: bool = False,
) -> pd.DataFrame:
    train = data["sales_train"].iloc[indices, :].copy()
    test = data["test"].set_index("ID").copy()
    shops = data["shops"].copy()
    items = data["items"].copy()
    item_cats = data["item_categories"].copy()

    # expand test set with same columns as train set so we can concatenate them row-wise
    if is_test:
        test["item_cnt_month"] = 0
        test["date_block_num"] = 34

        # merge train and test to build features
        dataset = pd.concat([train, test], axis=0)
    else:
        dataset = train.copy()

    # extract month and year
    dataset["year"] = dataset["date_block_num"].apply(lambda x: ((x // 12) + 2013))
    dataset["month"] = dataset["date_block_num"].apply(lambda x: (x % 12))

    # extract city name and shop type
    shops = extract_shop_features(df=dataset, shops=shops)

    shops = extract_cluster_feature(df=shops, cluster_name="shop_cluster", n_clusters=3)

    # extract category group/subgroup and fix some categories
    item_cats = extract_item_category_features(item_cats=item_cats)

    # extract first month sale, min item price, max item price
    items = extract_item_features(df=dataset, items=items)

    # items = extract_cluster_feature(
    #     df=items.merge(item_cats, on="item_category_id", how="left"),
    #     cluster_name="item_cluster",
    #     n_clusters=3,
    # )

    # extract lags of item_cnt_month
    dataset = extract_lagged_features(
        df=dataset, lags=[1, 2, 3, 4, 5, 6, 12], col_name="item_cnt_month"
    )

    # extract rolling window features
    dataset = extract_rolling_window_features(
        df=dataset, windows=[3], col_name="item_cnt_month", agg_func="mean"
    )

    # merge dataframes
    dataset = (
        dataset.merge(items, on="item_id", how="left")
        .merge(item_cats, on="item_category_id", how="left")
        .merge(shops, on="shop_id", how="left")
    )

    # drop unecessary column
    dataset.drop(columns=["item_price"], inplace=True)

    # cast dtypes
    dataset = reduce_mem_usage(dataset)

    if not is_test:
        return dataset
    else:
        return dataset[dataset["date_block_num"] == 34].drop(columns=["item_cnt_month"])
